// Copyright 2017 Fortanix, Inc. All rights reserved.
//
// This example demonstrates how to connect to the SDKMS server as an
// application using an API key and perform encryption and decryption
// using public key cryptography.

package com.fortanix.sdkms.v1.examples;

import java.io.IOException;
import java.util.Base64;
import java.util.List;
import org.apache.commons.cli.*;
import org.apache.commons.io.IOUtils;

// API library imports.
import com.fortanix.sdkms.v1.ApiClient;
import com.fortanix.sdkms.v1.ApiException;
import com.fortanix.sdkms.v1.Configuration;
import com.fortanix.sdkms.v1.api.AuthenticationApi;
import com.fortanix.sdkms.v1.api.EncryptionAndDecryptionApi;
import com.fortanix.sdkms.v1.api.SecurityObjectsApi;
import com.fortanix.sdkms.v1.auth.ApiKeyAuth;
import com.fortanix.sdkms.v1.model.AuthResponse;
import com.fortanix.sdkms.v1.model.DecryptRequest;
import com.fortanix.sdkms.v1.model.DecryptResponse;
import com.fortanix.sdkms.v1.model.EncryptRequest;
import com.fortanix.sdkms.v1.model.EncryptResponse;
import com.fortanix.sdkms.v1.model.KeyObject;
import com.fortanix.sdkms.v1.model.ObjectType;

// For convenience, you might want to just use the following imports instead:
//import com.fortanix.sdkms.v1.*;
//import com.fortanix.sdkms.v1.api.*;
//import com.fortanix.sdkms.v1.auth.*;
//import com.fortanix.sdkms.v1.models.*;

public class PubKeyCrypto {

    private enum CryptoOp {
        ENCRYPT,
        DECRYPT
    }

    private boolean debug;
    private boolean debug_tls;
    private String appId;
    private String server;
    private String truststore;
    private String apiKey;
    private String keyId;
    private String keyName;
    private String bearerToken;
    private CryptoOp direction;

    public void connect() {
        ApiClient client = new ApiClient();
        // Set the name of the server to talk to.
        client.setBasePath(server);

        // This optionally enables verbose logging in the API library.
        client.setDebugging(debug);

        // The default ApiClient (and its configured authorization) will be
        // used for constructing the specific API objects, such as
        // EncryptionAndDecryptionApi and SecurityObjectsApi.
        Configuration.setDefaultApiClient(client);

        // If you need a trust store for the server's certificate, you can
        // configure it here or via -Djavax.net.ssl.trustStore= on the
        // java command line.
        if (truststore != null) {
            System.setProperty("javax.net.ssl.trustStore", truststore);
        }

        // This optionally enables very verbose logging in the Java network
        // libraries.
        if (debug_tls) {
            System.setProperty("javax.net.debug", "all");
        }

        // When authenticating as an application, the API Key functions as
        // the entire HTTP basic auth token.
        client.setBasicAuthString(apiKey);

        // Acquire a bearer token to use for other APIs.
        try {
            AuthResponse response = new AuthenticationApi().authorize();
            bearerToken = response.getAccessToken();
            if (debug) {
                System.err.printf("Received Bearer token %s\n", bearerToken);
            }

            // Configure the client library to use the bearer token.
            ApiKeyAuth bearerAuth =
                (ApiKeyAuth) client.getAuthentication("bearerToken");
            bearerAuth.setApiKey(bearerToken);
            bearerAuth.setApiKeyPrefix("Bearer");
        } catch (ApiException e) {
            System.err.println("Unable to authenticate: " + e.getMessage());
            System.exit(1);
        }
    }

    public KeyObject findKey() throws ApiException {
        if (keyName != null) {
            // This filters by key name. Since names are unique, this will
            // return either 0 or 1 object.
            List<KeyObject> keys =
                new SecurityObjectsApi().getSecurityObjects(keyName, null, null);
            if (keys.size() == 0) {
                throw new ApiException("No key found with name " + keyName);
            } else {
                return keys.get(0);
            }
        } else {
            // This returns the security object with the given ID.
            return new SecurityObjectsApi().getSecurityObject(keyId);
        }
    }

    public void performCrypto() throws ApiException, IOException {
        // Find a key by name or ID. You could skip this step if you already
        // know the ID of the key you wish to encrypt or decrypt with and
        // the key type.
        KeyObject key = findKey();
        if (key.getObjType() != ObjectType.RSA) {
            throw new ApiException("Key of wrong type for public key crypto: " +
                                   key.getObjType().toString());
        }

        // This program gets the input data from stdin.
        byte[] input = IOUtils.toByteArray(System.in);

        byte[] output;
        
        if (direction == CryptoOp.ENCRYPT) {
            // For public key encryption, the plaintext and algorithm must
            // be supplied in the request body. The other properties of
            // the EncryptRequest model object are for symmetric encryption
            // and are not needed for public key encryption.
            EncryptRequest request = new EncryptRequest()
                .plain(input)
                .alg(ObjectType.RSA);
            EncryptResponse response =
                // The ID of the key is provided in the API call.
                new EncryptionAndDecryptionApi().encrypt(key.getKid(), request);
            output = response.getCipher();
        } else {
            // For public key decryption, the ciphertext and algorithm must
            // be supplied in the request body. The other properties of
            // the EncryptRequest model object are for symmetric encryption
            // and are not needed for public key decryption.
            DecryptRequest request = new DecryptRequest()
                .cipher(input)
                .alg(ObjectType.RSA);
            DecryptResponse response =
                // The ID of the key is provided in the API call.
                new EncryptionAndDecryptionApi().decrypt(key.getKid(), request);
            output = response.getPlain();
        }
        // And then we output to stdout.
        IOUtils.write(output, System.out);
    }

    public void logout() {
        if (bearerToken != null) {
            // It is a good idea to terminate the session when you are done
            // using it. This minimizes the window of time in which an attacker
            // could steal your bearer token and use it.
            try {
                new AuthenticationApi().terminate();
            } catch (ApiException e) {
                System.err.println("Error logging out: " + e.getMessage());
            }
            bearerToken = null;
        }
    }

    protected void usage(Options options) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp("Crypto",
                            "Encrypts or decrypts data using keys in " +
                            "an SDKMS server. Input is read from stdin " +
                            "and written to stdout", options, "");
    }

    protected void help(String msg) {
        System.err.println(msg);
        System.err.println("Run PubKeyCrypto --help for usage information");
    }

    public void parseCmdline(String[] args) {
        Options options = new Options();
        options.addOption(Option.builder("s")
                          .longOpt("server")
                          .hasArg()
                          .type(String.class)
                          .desc("SDKMS server endpoint (e.g. " +
                                "https://apps.smartkey.io)")
                          .build());

        options.addOption(Option.builder("k")
                          .longOpt("api-key")
                          .required(true)
                          .hasArg()
                          .type(String.class)
                          .desc("API key for this application")
                          .build());

        options.addOption(Option.builder("K")
                          .longOpt("key-id")
                          .hasArg()
                          .type(String.class)
                          .desc("Key ID of the cryptographic key to use")
                          .build());

        options.addOption(Option.builder("n")
                          .longOpt("key-name")
                          .hasArg()
                          .type(String.class)
                          .desc("Key name of the cryptographic key to use")
                          .build());

        options.addOption(Option.builder("e")
                          .longOpt("encrypt")
                          .desc("Perform encryption")
                          .build());

        options.addOption(Option.builder("E")
                          .longOpt("decrypt")
                          .desc("Perform decryption")
                          .build());

        options.addOption(Option.builder("t")
                          .longOpt("truststore")
                          .hasArg()
                          .type(String.class)
                          .desc("Java truststore containing server certificate")
                          .build());

        options.addOption(Option.builder("d")
                          .longOpt("debug")
                          .desc("Enable client API library debugging")
                          .build());

        options.addOption(Option.builder("D")
                          .longOpt("debug-tls")
                          .desc("Enable Java TLS debugging")
                          .build());

        options.addOption(Option.builder("h")
                          .longOpt("help")
                          .desc("Get usage information and exit")
                          .build());

        CommandLineParser parser = new DefaultParser();
        try {
            CommandLine cmdline = parser.parse(options, args);
            if (cmdline.hasOption("help")) {
                usage(options);
                System.exit(0);
            }

            debug = cmdline.hasOption("debug");
            debug_tls = cmdline.hasOption("debug-tls");
            appId = cmdline.getOptionValue("app-id");
            apiKey = cmdline.getOptionValue("api-key");

            if (cmdline.hasOption("server")) {
                server = cmdline.getOptionValue("server");
            } else {
                server = "https://apps.smartkey.io";
            }

            if (cmdline.hasOption("truststore")) {
                truststore = cmdline.getOptionValue("truststore");
            }

            if ((!cmdline.hasOption("encrypt") &&
                 !cmdline.hasOption("decrypt")) ||
                (cmdline.hasOption("encrypt") &&
                 cmdline.hasOption("decrypt"))) {
                help("You must specify either --encrypt or --decrypt");
                System.exit(1);
            }

            if (cmdline.hasOption("encrypt")) {
                direction = CryptoOp.ENCRYPT;
            } else {
                direction = CryptoOp.DECRYPT;
            }

            if (!cmdline.hasOption("key-id") &&
                !cmdline.hasOption("key-name")) {
                help("You must specify an encryption key with --key-id " +
                     "or --key-name");
                System.exit(1);
            }

            if (cmdline.hasOption("key-id") && cmdline.hasOption("key-name")) {
                System.err.println("Both --key-id and --key-name were " +
                                   "provided");
                System.err.println("Using the key name and ignoring id");
            }

            if (cmdline.hasOption("key-id")) {
                keyId = cmdline.getOptionValue("key-id");
            } else if (cmdline.hasOption("key-name")) {
                keyName = cmdline.getOptionValue("key-name");
            }

        } catch (ParseException e) {
            usage(options);
            System.err.println("Command-line parsing failed: " +
                               e.getMessage());
            System.exit(1);
        }
    }

    public static void main(String[] args) throws ApiException {
        int status = 0;

        PubKeyCrypto crypto = new PubKeyCrypto();
        crypto.parseCmdline(args);
        crypto.connect();
        try {
            crypto.performCrypto();
        } catch (ApiException | IOException e) {
            System.err.println("Crypto failed: " + e.getMessage());
            status = 1;
        } finally {
            crypto.logout();
        }
        System.exit(status);
    }
}
